import logging
import os


class DIDFormatter(logging.Formatter):
    """
    Need a customer formatter to allow for logging with request ids that vary.
    Normally log messages are "level instance component request_id msg" and
    request_id gets set by initialize_logging but we need a handler that'll let
    us pass in the request id and have that embedded in the log message
    """

    def format(self, record: logging.LogRecord) -> str:
        """
        Format record with request id if present, otherwise assume None

        :param record: LogRecord
        :return: formatted log message
        """

        if hasattr(record, "requestId"):
            return super().format(record)
        else:
            setattr(record, "requestId", None)
            return super().format(record)


def initialize_root_logger(did_scheme: str):
    """
    Get a logger and initialize it so that it outputs the correct format
    :param did_scheme: The scheme name to identify his did finder in log messages.
    :return: logger with correct formatting that outputs to console
    """

    log = logging.getLogger()
    instance = os.environ.get('INSTANCE_NAME', 'Unknown')
    formatter = DIDFormatter('%(levelname)s ' +
                             f"{instance} {did_scheme}_did_finder " +
                             '%(requestId)s %(message)s')
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(logging.INFO)
    log.addHandler(handler)
    log.setLevel(logging.INFO)
    return log



import argparse
from datetime import datetime
import json
import logging
import time
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional
import sys

import pika
from make_it_sync import make_sync

from servicex_did_finder_lib.did_summary import DIDSummary
from servicex_did_finder_lib.did_logging import initialize_root_logger
from servicex_did_finder_lib.util_uri import parse_did_uri
from .servicex_adaptor import ServiceXAdapter

# The type for the callback method to handle DID's, supplied by the user.
UserDIDHandler = Callable[[str, Dict[str, Any]], AsyncGenerator[Dict[str, Any], None]]

# Given name, build the RabbitMQ queue name by appending this.
# This is backed into how ServiceX works - do not change unless it
# is also changed in the ServiceX_App
QUEUE_NAME_POSTFIX = '_did_requests'

# Easy to use local logger
__logging = logging.getLogger(__name__)
__logging.addHandler(logging.NullHandler())


class _accumulator:
    'Track or cache files depending on the mode we are operating in'

    def __init__(self, sx: ServiceXAdapter, sum: DIDSummary, hold_till_end: bool):
        self._servicex = sx
        self._summary = sum
        self._hold_till_end = hold_till_end
        self._file_cache: List[Dict[str, Any]] = []

    def add(self, file_info: Dict[str, Any]):
        'Track and inject the file back into the system'
        if self._hold_till_end:
            self._file_cache.append(file_info)
        else:
            self._summary.add_file(file_info)
            if self._summary.file_count == 1:
                self._servicex.post_transform_start()
            self._servicex.put_file_add(file_info)

    def send_on(self, count):
        'Send the accumulated files'
        if self._hold_till_end:
            self._hold_till_end = False
            files = sorted(self._file_cache, key=lambda x: x['paths'])
            for file_info in files[0:count]:
                self.add(file_info)

    def send_bulk(self, file_list: List[Dict[str, Any]]):
        'does a bulk put of files'
        for ifl in file_list:
            self._summary.add_file(ifl)
        self._servicex.put_file_add_bulk(file_list)
        self._servicex.post_transform_start()


async def run_file_fetch_loop(did: str, servicex: ServiceXAdapter, info: Dict[str, Any],
                              user_callback: UserDIDHandler):
    start_time = datetime.now()

    summary = DIDSummary(did)
    did_info = parse_did_uri(did)
    hold_till_end = did_info.file_count != -1
    acc = _accumulator(servicex, summary, hold_till_end)

    try:
        async for file_info in user_callback(did_info.did, info):
            if type(file_info) is dict:
                acc.add(file_info)
            else:
                acc.send_bulk(file_info)

    except Exception:
        if did_info.get_mode == 'all':
            raise

    # If we've been holding onto any files, we need to send them now.
    acc.send_on(did_info.file_count)

    # Simple error checking and reporting
    if summary.file_count == 0:
        servicex.post_status_update(f'DID Finder found zero files for dataset {did}',
                                    severity='fatal')

    elapsed_time = int((datetime.now()-start_time).total_seconds())
    servicex.put_fileset_complete(
        {
            "files": summary.file_count,
            "files-skipped": summary.files_skipped,
            "total-events": summary.total_events,
            "total-bytes": summary.total_bytes,
            "elapsed-time": elapsed_time,
        }
    )

    servicex.post_status_update(f'Completed load of files in {elapsed_time} seconds')


def rabbit_mq_callback(user_callback: UserDIDHandler, channel, method, properties, body,
                       file_prefix=None):
    '''rabbit_mq_callback Respond to RabbitMQ Message

    When a request to resolve a DID comes into the DID finder, we
    respond with this callback. This callback will remain active
    until the request has been completed satisfied (so for some
    DID finders, this could be a fairly long time.)

    Args:
        channel ([type]): RabbitMQ channel
        method ([type]): Delivery method
        properties ([type]): Properties of the message
        body ([type]): The body (json for us) of the message
        file_prefix([str]): Prefix to put in front of file paths to enable use of Cache service
    '''
    request_id = None  # set this in case we get an exception while loading request
    try:
        # Unpack the message. Really bad if we fail up here!
        did_request = json.loads(body)
        did = did_request['did']
        request_id = did_request['request_id']
        __logging.info(f'Received DID request {did_request}', extra={'requestId': request_id})
        servicex = ServiceXAdapter(did_request['service-endpoint'], file_prefix)
        servicex.post_status_update("DID Request received")

        info = {
            'request-id': request_id,
        }

        # Process the request and resolve the DID
        try:
            make_sync(run_file_fetch_loop)(did, servicex, info, user_callback)

        except Exception as e:
            _, exec_value, _ = sys.exc_info()
            __logging.exception('DID Request Failed', extra={'requestId': request_id})
            servicex.post_status_update(f'DID Request Failed for id {request_id}: '
                                        f'{str(e)} - {exec_value}',
                                        severity='fatal')
            raise

    except Exception as e:
        __logging.exception(f'DID request failed {str(e)}', extra={'requestId': request_id})

    finally:
        channel.basic_ack(delivery_tag=method.delivery_tag)


def init_rabbit_mq(user_callback: UserDIDHandler,
                   rabbitmq_url: str, queue_name: str, retries: int,
                   retry_interval: float,
                   file_prefix: str = None):  # type: ignore
    rabbitmq = None
    retry_count = 0

    while not rabbitmq:
        try:
            rabbitmq = pika.BlockingConnection(pika.URLParameters(rabbitmq_url))
            _channel = rabbitmq.channel()
            _channel.queue_declare(queue=queue_name)

            __logging.info("Connected to RabbitMQ. Ready to start consuming requests")

            _channel.basic_consume(queue=queue_name,
                                   auto_ack=False,
                                   on_message_callback=lambda c, m, p, b:
                                   rabbit_mq_callback(user_callback, c, m, p, b, file_prefix))
            _channel.start_consuming()

        except pika.exceptions.AMQPConnectionError:  # type: ignore
            rabbitmq = None
            retry_count += 1
            if retry_count <= retries:
                __logging.exception(f'Failed to connect to RabbitMQ at {rabbitmq_url} '
                                    f'(try #{retry_count}). Waiting {retry_interval} seconds '
                                    'before trying again')
                time.sleep(retry_interval)
            else:
                __logging.exception(f'Failed to connect to RabbitMQ. Giving Up after {retry_count}'
                                    ' tries')
                raise


def add_did_finder_cnd_arguments(parser: argparse.ArgumentParser):
    '''add_did_finder_cnd_arguments Add required arguments to a parser

    If you need to parse command line arguments for some special configuration, create your
    own argument parser, and call this function to make sure the arguments needed
    for running the back-end communication are filled in properly.

    Then pass the results of the parsing to the `start_did_finder` method.

    Args:
        parser (argparse.ArgumentParser): The argument parser. Arguments needed for the
                                          did finder/servicex communication will be added.
    '''
    parser.add_argument('--rabbit-uri', dest="rabbit_uri", action='store', required=True)
    parser.add_argument('--prefix', dest="prefix", action='store', required=False,
                        default="",
                        help='Prefix to add to use a caching proxy for URIs')


def start_did_finder(did_finder_name: str,
                     callback: UserDIDHandler,
                     parsed_args: Optional[argparse.Namespace] = None):
    '''start_did_finder Start the DID finder

    Top level method that starts the DID finder, hooking it up to rabbitmq queues, etc.,
    and sets up the callback to be called each time ServiceX wants to render a DID into
    files.

    Once called this will not return unless it totally fails to connect to the rabbit
    mq server (which must have been specified on the command line that started this).
    If it can't return, then it will raise the appropriate exception.

    Args:
        did_name (str): Name of the DID finder (baked into the rabbit mq name)
        callback (UserDIDHandler): Callback to handle the DID rendering requests
        parser (Optional[argparse.Namespace], optional): If you need to parse your own
                                                         an arg parser, make sure to call
                                                         add_did_finder_cnd_arguments, run the
                                                         parser and pass in the resulting
                                                         parsed arguments.
                                                         Defaults to None (automatically
                                                         parses)
                                                         command line arguments, create
    '''
    # Setup command line parsing
    if parsed_args is None:
        parser = argparse.ArgumentParser()
        add_did_finder_cnd_arguments(parser)
        parsed_args = parser.parse_args()

    # Initialize the root logger
    initialize_root_logger(did_finder_name)

    # Start up rabbit mq and also callbacks
    init_rabbit_mq(callback,
                   parsed_args.rabbit_uri,
                   f'{did_finder_name}{QUEUE_NAME_POSTFIX}',
                   retries=12,
                   retry_interval=10)



from typing import Dict, List
import urllib


class ParsedDIDInfo:
    def __init__(self, did: str, get_mode: str, file_count: int):
        self.did = did
        self.get_mode = get_mode
        self.file_count = file_count

    # The did to pass into the library
    did: str

    # Mode to get the files (default 'all')
    get_mode: str

    # Number of files to fetch (default '-1')
    file_count: int


def parse_did_uri(uri: str) -> ParsedDIDInfo:
    '''Parse the uri that is given to us from ServiceX, pulling out
    the components we care about, and keeping the DID that needs to
    be passed down.

    URI arguments parsed:

    * `files` - Number of files to fetch (default is all)
    * `get` - Mode to get the files (default is 'all'). Only "available" is also supported.

    Args:
        uri (str): DID from ServiceX

    Returns:
        ParsedDIDInfo: The URI parsed into parts
    '''
    info = urllib.parse.urlparse(uri)  # type: ignore

    params = urllib.parse.parse_qs(info.query)  # type: ignore
    get_string = 'all' if 'get' not in params else params['get'][-1]
    file_count = -1 if 'files' not in params else int(params['files'][0])

    if get_string not in ['all', 'available']:
        raise ValueError('Bad value for "get" string in DID - must be "all" or "available", not '
                         f'"{get_string}"')

    for k in ['get', 'files']:
        if k in params:
            del params[k]

    def unwind_params(ps: Dict[str, List[str]]):
        for k, values in ps.items():
            for v in values:
                yield k, v

    new_query = "&".join(f'{k}={v}' for k, v in unwind_params(params))
    if len(new_query) > 0:
        new_query = "?" + new_query

    return ParsedDIDInfo(info._replace(query="").geturl() + new_query, get_string, file_count)



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
from datetime import datetime
import requests
import logging


MAX_RETRIES = 3


class ServiceXAdapter:
    def __init__(self, endpoint, file_prefix=None):
        self.endpoint = endpoint
        self.file_prefix = file_prefix

        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.NullHandler())

    def post_status_update(self, status_msg, severity="info"):
        success = False
        attempts = 0
        while not success and attempts < MAX_RETRIES:
            try:
                requests.post(self.endpoint + "/status", data={
                    "timestamp": datetime.now().isoformat(),
                    "source": "DID Finder",
                    "severity": severity,
                    "info": status_msg
                })
                success = True
            except requests.exceptions.ConnectionError:
                self.logger.exception(f'Connection error to ServiceX App. Will retry '
                                      f'(try {attempts} out of {MAX_RETRIES}')
                attempts += 1
        if not success:
            self.logger.error(f'After {attempts} tries, failed to send ServiceX App a status '
                              f'message: {str(status_msg)} - Ignoring error.')

    def _prefix_file(self, file_path):
        return file_path if not self.file_prefix else self.file_prefix+file_path

    def _create_json(self, file_info):
        return {
            "timestamp": datetime.now().isoformat(),
            "paths": [self._prefix_file(fp) for fp in file_info['paths']],
            'adler32': file_info['adler32'],
            'file_size': file_info['file_size'],
            'file_events': file_info['file_events']
        }

    def put_file_add(self, file_info):
        success = False
        attempts = 0
        while not success and attempts < MAX_RETRIES:
            try:
                mesg = self._create_json(file_info)
                requests.put(self.endpoint + "/files", json=mesg)
                self.logger.info(f"Metric: {json.dumps(mesg)}")
                success = True
            except requests.exceptions.ConnectionError:
                self.logger.exception(f'Connection error to ServiceX App. Will retry '
                                      f'(try {attempts} out of {MAX_RETRIES}')
                attempts += 1
        if not success:
            self.logger.error(f'After {attempts} tries, failed to send ServiceX App a put_file '
                              f'message: {str(file_info)} - Ignoring error.')

    def put_file_add_bulk(self, file_list, chunk_length=30):
        # we first chunk up file_list as it can be very large in
        # case there are a lot of replicas and a lot of files.
        chunks = [file_list[i:i + chunk_length] for i in range(0, len(file_list), chunk_length)]
        for chunk in chunks:
            success = False
            attempts = 0
            mesg = []
            for fi in chunk:
                mesg.append(self._create_json(fi))
            while not success and attempts < MAX_RETRIES:
                try:
                    requests.put(self.endpoint + "/files", json=mesg)
                    self.logger.info(f"Metric: {json.dumps(mesg)}")
                    success = True
                except requests.exceptions.ConnectionError:
                    self.logger.exception(f'Connection error to ServiceX App. Will retry '
                                          f'(try {attempts} out of {MAX_RETRIES}')
                    attempts += 1
            if not success:
                self.logger.error(f'After {attempts} tries, failed to send ServiceX App '
                                  f'a put_file_bulk message: {mesg} - Ignoring error.')

    def post_transform_start(self):
        success = False
        attempts = 0
        while not success and attempts < MAX_RETRIES:
            try:
                requests.post(self.endpoint + "/start")
                success = True
            except requests.exceptions.ConnectionError:
                self.logger.exception(f'Connection error to ServiceX App. Will retry '
                                      f'(try {attempts} out of {MAX_RETRIES}')
                attempts += 1
        if not success:
            self.logger.error(f'After {attempts} tries, failed to send ServiceX App a  '
                              f'transform start message - Ignoring error.')

    def put_fileset_complete(self, summary):
        success = False
        attempts = 0
        while not success and attempts < MAX_RETRIES:
            try:
                requests.put(self.endpoint + "/complete", json=summary)
                success = True
            except requests.exceptions.ConnectionError:
                self.logger.exception(f'Connection error to ServiceX App. Will retry '
                                      f'(try {attempts} out of {MAX_RETRIES}')
                attempts += 1
        if not success:
            self.logger.error(f'After {attempts} tries, failed to send ServiceX App a put_file '
                              f'message: {str(summary)} - Ignoring error.')



# Copyright (c) 2019, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import Any, Dict


class DIDSummary:
    def __init__(self, did: str):
        '''__init__ Track files we are reporting back to the system

        Object that tracks statistics about the files we are injecting for processing
        into Servicex.

        Args:
            did (string): The DID of the dataset we are tracking
        '''
        self._did = did
        self._total_bytes = 0
        self._total_events = 0
        self._files = 0
        self._files_skipped = 0

    def __str__(self):
        return ("DID {} - {:.0f} Mb {} Events in {} files ({} skipped)".format(
            self._did,
            self._total_bytes / 1e6,
            self._total_events,
            self._files,
            self._files_skipped))

    @property
    def file_count(self) -> int:
        return self._files

    @property
    def files_skipped(self) -> int:
        return self._files_skipped

    @property
    def total_bytes(self) -> int:
        return self._total_bytes

    @property
    def total_events(self) -> int:
        return self._total_events

    def _accumulate(self, file_record: Dict[str, Any]):
        '''accumulate Update statistics counters

        Track total file size and events

        Args:
            file_record (Dict[str, Any]): Statistics reported back from loading
                                          a particular for.
        '''
        file_size_key = 'file_size' if 'file_size' in file_record else 'bytes'
        file_events_key = 'file_events' if 'file_events' in file_record else 'events'

        self._total_bytes += int(file_record[file_size_key] or 0)
        self._total_events += int(file_record[file_events_key] or 0)

    def add_file(self, file_record: Dict[str, Any]):
        '''add_file Update all stats with new file

        Count number of files and track data for the file

        Args:
            file_record (Dict[str, Any]): Statistics for a particular file
        '''
        self._files += 1
        self._accumulate(file_record)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['servicex_did_finder_lib']

package_data = \
{'': ['*']}

install_requires = \
['make-it-sync>=1.0.0,<2.0.0', 'pika==1.1.0', 'requests>=2.25.0,<3.0.0']

setup_kwargs = {
    'name': 'servicex-did-finder-lib',
    'version': '1.2',
    'description': 'ServiceX DID Library Routines',
    'long_description': None,
    'author': 'Gordon Watts',
    'author_email': 'gwatts@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
